@@ -1,16 +1,18 @@
 //! Store and manipulate Kafka messages.
 
-use std::ffi::{CStr, CString};
+use std::ffi::CStr;
 use std::fmt;
 use std::marker::PhantomData;
-use std::os::raw::c_void;
+use std::os::raw::{c_char, c_void};
 use std::ptr;
 use std::str;
+use std::sync::Arc;
 use std::time::SystemTime;
 
 use rdkafka_sys as rdsys;
 use rdkafka_sys::types::*;
 
+use crate::admin::NativeEvent;
 use crate::error::{IsError, KafkaError, KafkaResult};
 use crate::util::{self, millis_to_epoch, KafkaDrop, NativePtr};
 
@@ -71,15 +73,103 @@ pub trait Headers {
     /// Returns the number of contained headers.
     fn count(&self) -> usize;
 
-    /// Gets the specified header, where the first header corresponds to index
-    /// 0. If the index is out of bounds, returns `None`.
-    fn get(&self, idx: usize) -> Option<(&str, &[u8])>;
+    /// Gets the specified header, where the first header corresponds to
+    /// index 0.
+    ///
+    /// Panics if the index is out of bounds.
+    fn get(&self, idx: usize) -> Header<'_, &[u8]> {
+        self.try_get(idx).unwrap_or_else(|| {
+            panic!(
+                "headers index out of bounds: the count is {} but the index is {}",
+                self.count(),
+                idx,
+            )
+        })
+    }
 
-    /// Like [`Headers::get`], but the value of the header will be converted to
-    /// the specified type. If the conversion fails, returns an error.
-    fn get_as<V: FromBytes + ?Sized>(&self, idx: usize) -> Option<(&str, Result<&V, V::Error>)> {
-        self.get(idx)
-            .map(|(name, value)| (name, V::from_bytes(value)))
+    /// Like [`Headers::get`], but the value of the header will be converted
+    /// to the specified type.
+    ///
+    /// Panics if the index is out of bounds.
+    fn get_as<V>(&self, idx: usize) -> Result<Header<'_, &V>, V::Error>
+    where
+        V: FromBytes + ?Sized,
+    {
+        self.try_get_as(idx).unwrap_or_else(|| {
+            panic!(
+                "headers index out of bounds: the count is {} but the index is {}",
+                self.count(),
+                idx,
+            )
+        })
+    }
+
+    /// Like [`Headers::get`], but returns an option if the header is out of
+    /// bounds rather than panicking.
+    fn try_get(&self, idx: usize) -> Option<Header<'_, &[u8]>>;
+
+    /// Like [`Headers::get`], but returns an option if the header is out of
+    /// bounds rather than panicking.
+    fn try_get_as<V>(&self, idx: usize) -> Option<Result<Header<'_, &V>, V::Error>>
+    where
+        V: FromBytes + ?Sized,
+    {
+        self.try_get(idx).map(|header| header.parse())
+    }
+
+    /// Iterates over all headers in order.
+    fn iter(&self) -> HeadersIter<'_, Self>
+    where
+        Self: Sized,
+    {
+        HeadersIter {
+            headers: self,
+            index: 0,
+        }
+    }
+}
+
+/// A Kafka message header.
+#[derive(Debug, Clone, Eq, PartialEq, Ord, PartialOrd)]
+pub struct Header<'a, V> {
+    /// The header's key.
+    pub key: &'a str,
+    /// The header's value.
+    pub value: Option<V>,
+}
+
+impl<'a> Header<'a, &'a [u8]> {
+    fn parse<V>(&self) -> Result<Header<'a, &'a V>, V::Error>
+    where
+        V: FromBytes + ?Sized,
+    {
+        Ok(Header {
+            key: self.key,
+            value: self.value.map(V::from_bytes).transpose()?,
+        })
+    }
+}
+
+/// An iterator over [`Headers`].
+pub struct HeadersIter<'a, H> {
+    headers: &'a H,
+    index: usize,
+}
+
+impl<'a, H> Iterator for HeadersIter<'a, H>
+where
+    H: Headers,
+{
+    type Item = Header<'a, &'a [u8]>;
+
+    fn next(&mut self) -> Option<Header<'a, &'a [u8]>> {
+        if self.index < self.headers.count() {
+            let item = self.headers.get(self.index);
+            self.index += 1;
+            Some(item)
+        } else {
+            None
+        }
     }
 }
 
@@ -173,7 +263,7 @@ impl Headers for BorrowedHeaders {
         unsafe { rdsys::rd_kafka_header_cnt(self.as_native_ptr()) }
     }
 
-    fn get(&self, idx: usize) -> Option<(&str, &[u8])> {
+    fn try_get(&self, idx: usize) -> Option<Header<'_, &[u8]>> {
         let mut value_ptr = ptr::null();
         let mut name_ptr = ptr::null();
         let mut value_size = 0;
@@ -190,10 +280,11 @@ impl Headers for BorrowedHeaders {
             None
         } else {
             unsafe {
-                Some((
-                    CStr::from_ptr(name_ptr).to_str().unwrap(),
-                    util::ptr_to_slice(value_ptr, value_size),
-                ))
+                Some(Header {
+                    key: CStr::from_ptr(name_ptr).to_str().unwrap(),
+                    value: (!value_ptr.is_null())
+                        .then(|| util::ptr_to_slice(value_ptr, value_size)),
+                })
             }
         }
     }
@@ -217,17 +308,26 @@ impl Headers for BorrowedHeaders {
 /// [`detach`](BorrowedMessage::detach) method.
 pub struct BorrowedMessage<'a> {
     ptr: NativePtr<RDKafkaMessage>,
+    _event: Arc<NativeEvent>,
     _owner: PhantomData<&'a u8>,
 }
 
+// When using the Event API, messages must not be freed with rd_kafka_message_destroy
+unsafe extern "C" fn no_op(_: *mut RDKafkaMessage) {}
+
 unsafe impl KafkaDrop for RDKafkaMessage {
     const TYPE: &'static str = "message";
-    const DROP: unsafe extern "C" fn(*mut Self) = rdsys::rd_kafka_message_destroy;
+    const DROP: unsafe extern "C" fn(*mut Self) = no_op;
 }
 
 impl<'a> fmt::Debug for BorrowedMessage<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "Message {{ ptr: {:?} }}", self.ptr())
+        write!(
+            f,
+            "Message {{ ptr: {:?}, event_ptr: {:?} }}",
+            self.ptr(),
+            self._event.ptr()
+        )
     }
 }
 
@@ -238,14 +338,15 @@ impl<'a> BorrowedMessage<'a> {
     /// should only be used with messages coming from consumers. If the message
     /// contains an error, only the error is returned and the message structure
     /// is freed.
-    pub(crate) unsafe fn from_consumer<C>(
+    pub(crate) unsafe fn from_client<C>(
         ptr: NativePtr<RDKafkaMessage>,
-        _consumer: &'a C,
+        event: Arc<NativeEvent>,
+        _client: &'a C,
     ) -> KafkaResult<BorrowedMessage<'a>> {
         if ptr.err.is_error() {
             let err = match ptr.err {
                 rdsys::rd_kafka_resp_err_t::RD_KAFKA_RESP_ERR__PARTITION_EOF => {
-                    KafkaError::PartitionEOF((*ptr).partition)
+                    KafkaError::PartitionEOF(ptr.partition)
                 }
                 e => KafkaError::MessageConsumption(e.into()),
             };
@@ -253,22 +354,24 @@ impl<'a> BorrowedMessage<'a> {
         } else {
             Ok(BorrowedMessage {
                 ptr,
+                _event: event,
                 _owner: PhantomData,
             })
         }
     }
 
     /// Creates a new `BorrowedMessage` that wraps the native Kafka message
-    /// pointer returned by the delivery callback of a producer. The lifetime of
-    /// the message will be bound to the lifetime of the reference passed as
-    /// parameter. This method should only be used with messages coming from the
-    /// delivery callback. The message will not be freed in any circumstance.
-    pub(crate) unsafe fn from_dr_callback<O>(
+    /// pointer returned via the delivery report event. The lifetime of
+    /// the message will be bound to the lifetime of the client passed as
+    /// parameter.
+    pub(crate) unsafe fn from_dr_event<C>(
         ptr: *mut RDKafkaMessage,
-        _owner: &'a O,
+        event: Arc<NativeEvent>,
+        _client: &'a C,
     ) -> DeliveryResult<'a> {
         let borrowed_message = BorrowedMessage {
             ptr: NativePtr::from_ptr(ptr).unwrap(),
+            _event: event,
             _owner: PhantomData,
         };
         if (*ptr).err.is_error() {
@@ -322,20 +425,20 @@ impl<'a> Message for BorrowedMessage<'a> {
     type Headers = BorrowedHeaders;
 
     fn key(&self) -> Option<&[u8]> {
-        unsafe { util::ptr_to_opt_slice((*self.ptr).key, (*self.ptr).key_len) }
+        unsafe { util::ptr_to_opt_slice(self.ptr.key, self.ptr.key_len) }
     }
 
     fn payload(&self) -> Option<&[u8]> {
-        unsafe { util::ptr_to_opt_slice((*self.ptr).payload, (*self.ptr).len) }
+        unsafe { util::ptr_to_opt_slice(self.ptr.payload, self.ptr.len) }
     }
 
     unsafe fn payload_mut(&mut self) -> Option<&mut [u8]> {
-        util::ptr_to_opt_mut_slice((*self.ptr).payload, (*self.ptr).len)
+        util::ptr_to_opt_mut_slice(self.ptr.payload, self.ptr.len)
     }
 
     fn topic(&self) -> &str {
         unsafe {
-            CStr::from_ptr(rdsys::rd_kafka_topic_name((*self.ptr).rkt))
+            CStr::from_ptr(rdsys::rd_kafka_topic_name(self.ptr.rkt))
                 .to_str()
                 .expect("Topic name is not valid UTF-8")
         }
@@ -426,17 +529,28 @@ impl OwnedHeaders {
         }
     }
 
-    /// Adds a new header.
-    pub fn add<V: ToBytes + ?Sized>(self, name: &str, value: &V) -> OwnedHeaders {
-        let name_cstring = CString::new(name.to_owned()).unwrap();
-        let value_bytes = value.to_bytes();
+    /// Inserts a new header.
+    pub fn insert<V>(self, header: Header<'_, &V>) -> OwnedHeaders
+    where
+        V: ToBytes + ?Sized,
+    {
+        let (value_ptr, value_len) = match header.value {
+            None => (ptr::null_mut(), 0),
+            Some(value) => {
+                let value_bytes = value.to_bytes();
+                (
+                    value_bytes.as_ptr() as *mut c_void,
+                    value_bytes.len() as isize,
+                )
+            }
+        };
         let err = unsafe {
             rdsys::rd_kafka_header_add(
                 self.ptr(),
-                name_cstring.as_ptr(),
-                name_cstring.as_bytes().len() as isize,
-                value_bytes.as_ptr() as *mut c_void,
-                value_bytes.len() as isize,
+                header.key.as_ptr() as *const c_char,
+                header.key.as_bytes().len() as isize,
+                value_ptr,
+                value_len,
             )
         };
         // OwnedHeaders should always represent writable instances of RDKafkaHeaders
@@ -465,8 +579,8 @@ impl Headers for OwnedHeaders {
         unsafe { rdsys::rd_kafka_header_cnt(self.ptr()) }
     }
 
-    fn get(&self, idx: usize) -> Option<(&str, &[u8])> {
-        self.as_borrowed().get(idx)
+    fn try_get(&self, idx: usize) -> Option<Header<'_, &[u8]>> {
+        self.as_borrowed().try_get(idx)
     }
 }
 
@@ -517,6 +631,65 @@ impl OwnedMessage {
             headers,
         }
     }
+
+    /// Detaches the [`OwnedHeaders`] from this `OwnedMessage`.
+    pub fn detach_headers(&mut self) -> Option<OwnedHeaders> {
+        self.headers.take()
+    }
+
+    /// Replaces the [`OwnedHeaders`] on this `OwnedMessage`.
+    pub fn replace_headers(mut self, headers: Option<OwnedHeaders>) -> Self {
+        if let Some(headers) = headers {
+            self.headers.replace(headers);
+        } else {
+            self.headers = None;
+        }
+        self
+    }
+
+    /// Sets the payload for this `OwnedMessage`.
+    pub fn set_payload(mut self, payload: Option<Vec<u8>>) -> Self {
+        if let Some(payload) = payload {
+            self.payload.replace(payload);
+        } else {
+            self.payload = None;
+        }
+        self
+    }
+
+    /// Sets the key for this `OwnedMessage`.
+    pub fn set_key(mut self, key: Option<Vec<u8>>) -> Self {
+        if let Some(key) = key {
+            self.key.replace(key);
+        } else {
+            self.key = None;
+        }
+        self
+    }
+
+    /// Sets the topic for this `OwnedMessage`.
+    pub fn set_topic(mut self, topic: String) -> Self {
+        self.topic = topic;
+        self
+    }
+
+    /// Sets the timestamp for this `OwnedMessage`.
+    pub fn set_timestamp(mut self, timestamp: Timestamp) -> Self {
+        self.timestamp = timestamp;
+        self
+    }
+
+    /// Sets the partition for this `OwnedMessage`.
+    pub fn set_partition(mut self, partition: i32) -> Self {
+        self.partition = partition;
+        self
+    }
+
+    /// Sets the offset for this `OwnedMessage`.
+    pub fn set_offset(mut self, offset: i64) -> Self {
+        self.offset = offset;
+        self
+    }
 }
 
 impl Message for OwnedMessage {
@@ -648,24 +821,12 @@ impl ToBytes for () {
     }
 }
 
-// Implement to_bytes for arrays - https://github.com/rust-lang/rfcs/issues/1038
-macro_rules! array_impls {
-    ($($N:expr)+) => {
-        $(
-            impl ToBytes for [u8; $N] {
-                fn to_bytes(&self) -> &[u8] { self }
-            }
-         )+
+impl<const N: usize> ToBytes for [u8; N] {
+    fn to_bytes(&self) -> &[u8] {
+        self
     }
 }
 
-array_impls! {
-     0  1  2  3  4  5  6  7  8  9
-    10 11 12 13 14 15 16 17 18 19
-    20 21 22 23 24 25 26 27 28 29
-    30 31 32
-}
-
 #[cfg(test)]
 mod test {
     use super::*;
@@ -696,12 +857,27 @@ mod test {
     #[test]
     fn test_headers() {
         let owned = OwnedHeaders::new()
-            .add("key1", "value1")
-            .add("key2", "value2");
+            .insert(Header {
+                key: "key1",
+                value: Some("value1"),
+            })
+            .insert(Header {
+                key: "key2",
+                value: Some("value2"),
+            });
         assert_eq!(
             owned.get(0),
-            Some(("key1", &[118, 97, 108, 117, 101, 49][..]))
+            Header {
+                key: "key1",
+                value: Some(&[118, 97, 108, 117, 101, 49][..])
+            }
+        );
+        assert_eq!(
+            owned.get_as::<str>(1),
+            Ok(Header {
+                key: "key2",
+                value: Some("value2")
+            })
         );
-        assert_eq!(owned.get_as::<str>(1), Some(("key2", Ok("value2"))));
     }
 }
