@@ -3,6 +3,7 @@
 //! See the [`FutureProducer`] for details.
 // TODO: extend docs
 
+use std::error::Error;
 use std::future::Future;
 use std::marker::PhantomData;
 use std::pin::Pin;
@@ -10,19 +11,24 @@ use std::sync::Arc;
 use std::task::{Context, Poll};
 use std::time::{Duration, Instant};
 
-use futures::channel::oneshot;
-use futures::FutureExt;
+use futures_channel::oneshot;
+use futures_util::FutureExt;
 
-use crate::client::{Client, ClientContext, DefaultClientContext};
+use crate::client::{Client, ClientContext, DefaultClientContext, OAuthToken};
 use crate::config::{ClientConfig, FromClientConfig, FromClientConfigAndContext, RDKafkaLogLevel};
 use crate::consumer::ConsumerGroupMetadata;
 use crate::error::{KafkaError, KafkaResult, RDKafkaErrorCode};
 use crate::message::{Message, OwnedHeaders, OwnedMessage, Timestamp, ToBytes};
-use crate::producer::{BaseRecord, DeliveryResult, Producer, ProducerContext, ThreadedProducer};
+use crate::producer::{
+    BaseRecord, DeliveryResult, NoCustomPartitioner, Producer, ProducerContext, PurgeConfig,
+    ThreadedProducer,
+};
 use crate::statistics::Statistics;
 use crate::topic_partition_list::TopicPartitionList;
 use crate::util::{AsyncRuntime, DefaultRuntime, IntoOpaque, Timeout};
 
+use super::Partitioner;
+
 //
 // ********** FUTURE PRODUCER **********
 //
@@ -137,6 +143,8 @@ pub type OwnedDeliveryResult = Result<(i32, i64), (KafkaError, OwnedMessage)>;
 
 // Delegates all the methods calls to the wrapped context.
 impl<C: ClientContext + 'static> ClientContext for FutureProducerContext<C> {
+    const ENABLE_REFRESH_OAUTH_TOKEN: bool = C::ENABLE_REFRESH_OAUTH_TOKEN;
+
     fn log(&self, level: RDKafkaLogLevel, fac: &str, log_message: &str) {
         self.wrapped_context.log(level, fac, log_message);
     }
@@ -152,9 +160,21 @@ impl<C: ClientContext + 'static> ClientContext for FutureProducerContext<C> {
     fn error(&self, error: KafkaError, reason: &str) {
         self.wrapped_context.error(error, reason);
     }
+
+    fn generate_oauth_token(
+        &self,
+        oauthbearer_config: Option<&str>,
+    ) -> Result<OAuthToken, Box<dyn Error>> {
+        self.wrapped_context
+            .generate_oauth_token(oauthbearer_config)
+    }
 }
 
-impl<C: ClientContext + 'static> ProducerContext for FutureProducerContext<C> {
+impl<C, Part> ProducerContext<Part> for FutureProducerContext<C>
+where
+    C: ClientContext + 'static,
+    Part: Partitioner,
+{
     type DeliveryOpaque = Box<oneshot::Sender<OwnedDeliveryResult>>;
 
     fn delivery(
@@ -182,11 +202,12 @@ impl<C: ClientContext + 'static> ProducerContext for FutureProducerContext<C> {
 /// underlying producer. The internal polling thread will be terminated when the
 /// `FutureProducer` goes out of scope.
 #[must_use = "Producer polling thread will stop immediately if unused"]
-pub struct FutureProducer<C = DefaultClientContext, R = DefaultRuntime>
+pub struct FutureProducer<C = DefaultClientContext, R = DefaultRuntime, Part = NoCustomPartitioner>
 where
+    Part: Partitioner,
     C: ClientContext + 'static,
 {
-    producer: Arc<ThreadedProducer<FutureProducerContext<C>>>,
+    producer: Arc<ThreadedProducer<FutureProducerContext<C>, Part>>,
     _runtime: PhantomData<R>,
 }
 
@@ -202,8 +223,11 @@ where
     }
 }
 
-impl FromClientConfig for FutureProducer {
-    fn from_config(config: &ClientConfig) -> KafkaResult<FutureProducer> {
+impl<R> FromClientConfig for FutureProducer<DefaultClientContext, R>
+where
+    R: AsyncRuntime,
+{
+    fn from_config(config: &ClientConfig) -> KafkaResult<FutureProducer<DefaultClientContext, R>> {
         FutureProducer::from_config_and_context(config, DefaultClientContext)
     }
 }
@@ -322,6 +346,7 @@ where
 
     /// Like [`FutureProducer::send`], but if enqueuing fails, an error will be
     /// returned immediately, alongside the [`FutureRecord`] provided.
+    #[allow(clippy::result_large_err)]
     pub fn send_result<'a, K, P>(
         &self,
         record: FutureRecord<'a, K, P>,
@@ -347,17 +372,22 @@ where
     }
 }
 
-impl<C, R> Producer<FutureProducerContext<C>> for FutureProducer<C, R>
+impl<C, R, Part> Producer<FutureProducerContext<C>, Part> for FutureProducer<C, R, Part>
 where
     C: ClientContext + 'static,
     R: AsyncRuntime,
+    Part: Partitioner,
 {
     fn client(&self) -> &Client<FutureProducerContext<C>> {
         self.producer.client()
     }
 
-    fn flush<T: Into<Timeout>>(&self, timeout: T) {
-        self.producer.flush(timeout);
+    fn flush<T: Into<Timeout>>(&self, timeout: T) -> KafkaResult<()> {
+        self.producer.flush(timeout)
+    }
+
+    fn purge(&self, flags: PurgeConfig) {
+        self.producer.purge(flags)
     }
 
     fn in_flight_count(&self) -> i32 {
@@ -401,7 +431,7 @@ mod tests {
     struct TestContext;
 
     impl ClientContext for TestContext {}
-    impl ProducerContext for TestContext {
+    impl ProducerContext<NoCustomPartitioner> for TestContext {
         type DeliveryOpaque = Box<i32>;
 
         fn delivery(&self, _: &DeliveryResult<'_>, _: Self::DeliveryOpaque) {
@@ -413,6 +443,7 @@ mod tests {
     #[test]
     fn test_future_producer_clone() {
         let producer = ClientConfig::new().create::<FutureProducer>().unwrap();
+        #[allow(clippy::redundant_clone)]
         let _producer_clone = producer.clone();
     }
 
@@ -423,6 +454,7 @@ mod tests {
         let producer = ClientConfig::new()
             .create_with_context::<_, FutureProducer<TestContext>>(test_context)
             .unwrap();
+        #[allow(clippy::redundant_clone)]
         let _producer_clone = producer.clone();
     }
 }
