@@ -17,12 +17,12 @@ use rdkafka::producer::{FutureProducer, FutureRecord};
 use rdkafka::statistics::Statistics;
 use rdkafka::TopicPartitionList;
 
-pub fn rand_test_topic() -> String {
+pub fn rand_test_topic(test_name: &str) -> String {
     let id = rand::thread_rng()
         .gen_ascii_chars()
         .take(10)
         .collect::<String>();
-    format!("__test_{}", id)
+    format!("__{}_{}", test_name, id)
 }
 
 pub fn rand_test_group() -> String {
@@ -67,9 +67,7 @@ pub fn get_broker_version() -> KafkaVersion {
             panic!("KAFKA_VERSION env var contained non-unicode characters")
         }
         // If the environment variable is unset, assume we're running the latest version.
-        Err(VarError::NotPresent) => {
-            KafkaVersion(std::u32::MAX, std::u32::MAX, std::u32::MAX, std::u32::MAX)
-        }
+        Err(VarError::NotPresent) => KafkaVersion(u32::MAX, u32::MAX, u32::MAX, u32::MAX),
     }
 }
 
@@ -164,27 +162,6 @@ pub fn key_fn(id: i32) -> String {
     format!("Key {}", id)
 }
 
-#[cfg(test)]
-mod tests {
-    use super::*;
-
-    #[tokio::test]
-    async fn test_populate_topic() {
-        let topic_name = rand_test_topic();
-        let message_map = populate_topic(&topic_name, 100, &value_fn, &key_fn, Some(0), None).await;
-
-        let total_messages = message_map
-            .iter()
-            .filter(|&(&(partition, _), _)| partition == 0)
-            .count();
-        assert_eq!(total_messages, 100);
-
-        let mut ids = message_map.iter().map(|(_, id)| *id).collect::<Vec<_>>();
-        ids.sort();
-        assert_eq!(ids, (0..100).collect::<Vec<_>>());
-    }
-}
-
 pub struct ConsumerTestContext {
     pub _n: i64, // Add data for memory access validation
 }
@@ -228,3 +205,24 @@ pub fn consumer_config(
 
     config
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[tokio::test]
+    async fn test_populate_topic() {
+        let topic_name = rand_test_topic("test_populate_topic");
+        let message_map = populate_topic(&topic_name, 100, &value_fn, &key_fn, Some(0), None).await;
+
+        let total_messages = message_map
+            .iter()
+            .filter(|&(&(partition, _), _)| partition == 0)
+            .count();
+        assert_eq!(total_messages, 100);
+
+        let mut ids = message_map.values().copied().collect::<Vec<_>>();
+        ids.sort();
+        assert_eq!(ids, (0..100).collect::<Vec<_>>());
+    }
+}
